# Preprocessors

- [Introduction](#introduction)
- [Purpose](#purpose)
- [Features](#features)
- [How To Use](#how-to-use)

## Introduction
So you have a basic Drupal installation coupled with a nice theme, and
you want to customize the variables the templates receive. Normally,
you'd create a `THEMENAME.theme` file at the root of your theme and add
**hooks** to alter the `$variables` variable. But let's just
say you have a lot of preprocessing to do? This file can become huge,
messy and difficult to manage.

This is where the **Preprocessors** module comes in! This module uses the
powerful **Plugin API** to allow for preprocessing management in Plugins
rather than through procedural code in the theme.

## Purpose
The ultimate goal of this module is to better organize preprocessing of
templates in Drupal and to make it fit seamlessly into the new OOP
approach Drupal is taking.

## Features
- Preprocess template variables through the Plugin API.

## Usage
Normally, when attempting to preprocess the variables of a **node** for
example, you would create a function such as the following in your
THEME_NAME.theme file:

```php
/**
 * Implements hook_preprocess_HOOK() for page templates.
 */
function THEMENAME_preprocess_node(&$variables) {
    $variables['hello'] = 'world';
}
```

Often times, this file would become very large and difficult to truly
maintain, as some sites can have very specific preprocessing actions in
many entity/object types. This is where Preprocessors come in.

Preprocessor plugins have 4 possible properties:
- ID **(Required)**: The unique ID of the plugin. This is usually
`MODULE_NAME.CLASS_NAME` or `THEME_NAME.CLASS_NAME`.
- Template **(Required)**: The template that this plugin will applies
preprocessing for. This is equivalent to the value of `HOOK` in a
regular `hook_preprocess_HOOK` function.
- Weight **(Optional)**: The weight relative to other preprocessors that
act upon the same template. Lower value means earlier preprocessing.
**Defaults to 0.**
- Themes **(Optional)**: The Themes in which this plugin will act. If
the plugin is provided by a theme, this property should be omitted as it
will only act in the theme its defined in. Otherwise, in modules, it is
optional, and plugins will act in all themes if this is omitted.

There are two main ways to create **preprocessors**:
- Through a custom theme
- Through a module

Below, I will include examples of how to create functionality that does
exactly the same thing as the hook function above, but through Plugins.

### Adding preprocessors to a theme
In a theme, the first step would be to create a
`THEME_NAME.preprocessors.yml` at the root of the theme. The contents
of this file should be similar  to this:

```yaml
THEME_NAME.node_preprocessor:
  class: '\Drupal\THEME_NAME\Plugin\preprocessors\NodePreprocessor'
  template: node
  weight: 0
```

Here, you are declaring a plugin with the id
`THEME_NAME.node_preprocessor` and setting the path towards its class
that we will create shortly. The `template` key actually tells the
plugin which `template` we are preprocessing, and is equivalent to the
value you would replace `HOOK` with in any `hook_preprocess_HOOK()`
function.

Now, you can create a `NodePreprocessor.php` file.

File: `/THEME_NAME/src/Plugin/preprocessors/NodePreprocessor.php`

```php
<?php

namespace Drupal\THEME_NAME\Plugin\preprocessors;

use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorPluginBase;

/**
 * Provides a Preprocessor for nodes.
 */
final class NodePreprocessor extends PreprocessorPluginBase {

  /**
   * {@inheritdoc}
   */
  public function preprocess(&$variables, $hook, $info) {
    $variables['hello'] = 'world';
  }

}
```

And you're done! Clear your cache, and your class should now function
as a preprocessor for nodes! Beyond this, you can use Dependency
Injection to inherit some of Drupal's services, or use more advanced OOP
patterns to create any kind of relations between multiple preprocessors
you may have.

### Adding preprocessors to a module
Adding a preprocessor to a module is pretty much the same deal as
with a theme. The only difference is that you don't need a
`.preprocessors.yml` file at the root of your module. This is because
in modules, Plugins will discovered using the Class Annotation method.

Create the `NodePreprocessor.php` file.

File: `/MODULE_NAME/src/Plugin/preprocessors/NodePreprocessor.php`

```php
<?php

namespace Drupal\MODULE_NAME\Plugin\preprocessors;

use Drupal\preprocessors\PluginManager\PreprocessorsPluginManager\PreprocessorPluginBase;

/**
 * Provide plugin to alter variables for nodes.
 *
 * @noinspection
 *   AnnotationMissingUseInspection
 *
 * @Preprocessor(
 *   id = "preprocessors_example.node_preprocessor",
 *   template = "node",
 *   weight = 0,
 *   themes = {
 *      "YOUR_THEME"
 *   }
 * )
 *
 * @package Drupal\preprocessors_example\Plugin\Preprocessor
 */
final class NodePreprocessor extends PreprocessorPluginBase {

  /**
   * Add personal tweaks to variables in this function.
   *
   * {@inheritdoc}
   */
  public function preprocess(&$variables, $hook, $info): void {
    $variables['hello'] = 'world';
  }

}
```

In modules, you define the plugin's details in the class Annotation, as
opposed to the `.yml` file in themes. These values function the same way
, with the `themes` property being more useful in modules in case you
want your preprocessor to be limited to a selection of themes. If this
property is omitted, it will be applied in all themes.

Otherwise, the rest is the same and straightforward!
